#!/usr/bin/env perl

################################################################################
#
# Copyright (C) 2011 Sarah Eris Horsley Caffee
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
#
# Returns a new unused username for use on a *nix server.
#
################################################################################

use warnings;
use strict;
use English;
use File::Basename;

# main is a block.  NO GLOBALS!
{
my $base ="";
my $user;
my $debug = 0;
my $maxlen = 8;
my $test = 0;

foreach my $arg (@ARGV) {
    if ("-d" eq $arg) {
	$debug = 1;
    } elsif ($arg =~ /^-m/) {
	$arg =~ s/^..//;
	if ($arg =~ /^(\d+)$/) {
	    $maxlen = $1;
	    $debug and print "Setting maxlen to $maxlen\n";
	} else {
	    printf(STDERR "Invalid username length specified: $arg\n");
	}
    } elsif ($arg eq "-h") {
	usage();
	exit 0;
    } elsif ($arg eq "-t") {
	$test = 1;
    } else {
	$base = $arg;
    }
}

if (! $test) {
    $user = make_username($base, $maxlen);
    if (defined $user) {
	printf("%s\n", $user);
    } else {
	printf(STDERR "No username available\n");
    }
    exit 0;
} else {
    my $pwdfile = "passwd";
    my ($base, $t1, $t2); 
    my @bases = ("", # should use default "user" as base.  user and user1 taken
		 "eightchr",
		 "short",
		 "this *&% is not allowed &^$%&^m",
		 "123digits",
		 "this-is-a-domain.com",
		 '&$%#&$*&$%*&',
		 "testqwer", # testqwer taken
		 "testabc", # all testabc? taken
		 "testzx", # all testzx?? taken
		 "test", # all test??? taken
		 "t"	# all z????? taken though 10435
		 );
 
    foreach $base (@bases) {
	$t1 = time;
	$user = make_username($base, $maxlen, $pwdfile);
	$t2 = time;
	if (defined $user) {
	    printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);
	} else {
	    printf("%30s No username available\n", $base, $t2 - $t1);
	}
    }
    
    $maxlen = 2;
    $base = "z"; # all z? taken.  Should return no username.
    $t1 = time;
    $user = make_username($base, $maxlen, $pwdfile);
    $t2 = time;
    if (defined $user) {
	printf("%30s %-".$maxlen."s in %ds\n", $base, $user, $t2 - $t1);
    } else {
	printf("%30s No username available\n", $base, $t2 - $t1);
    }
    
}

}
exit 0;

################################################################################
sub make_username {
    # Arguments:
    # $base The string to use as the basis of the new username.
    #       Can be any string (such as an existing username or a domain name)
    #       But at most $maxlen characters will be used.  Invalid characters will
    #       be removed.
    #       If the first character would be a number, an 'a' is prepended to the name.
    #       If blank, the string "user" will be used.
    # $maxlen The maximum length of the username to create.  If not specified, 
    #       it defaults to 8. It must be 2 or more. (If it was 1 then there 
    #       would be no way to prepend a number to it without having a digit as 
    #       the first character of the new name.) 
    # $pwdfile The name of the passwd file to use for checking for new users.
    #       /etc/passwd by default.  This is really a debugging option.
    #
    # Returns:
    # A string containing the new username or undef if no username 
    # could be generated  (unlikely)

    my ($base, $maxlen, $pwdfile) = @_;

    if ($base eq "") {
	$base = "user";
    }

    ! defined $pwdfile and $pwdfile = "/etc/passwd";

    if (! defined $maxlen) {
	$maxlen = 8;
    } else {
	if ($maxlen !~ /^\d+$/) {
	    printf(STDERR "Error: non-numeric username length specified.\n");
	    return undef;
	}
    }

    if (! -e $pwdfile) {
	printf(STDERR "Error: non-existant passwd file specified: $pwdfile\n");
	return undef;
    }

    # Remove non-alpha-numeric characters and make sure the first character 
    # is a letter.
    $base =~ s/[^A-Za-z0-9]//g;
    if ($base !~ /^[a-z]/ ) { 
        $base = "a".$base;
    }

    # Use the specified parameter as the base of the new username.
    # Print it into a string of exactly $maxlen characters.
    # Print the number into the last portion of the new username.
    # Delete spaces (in case the passed username candidate was short).
    # Churn until we find an unused name.
    my $i=1;
    my $numstr;
    my $newuser = sprintf("%.".$maxlen."s", $base);
    while ( (user_exists($newuser, $pwdfile)) && ($i < 10**($maxlen-1)) ) { 
	$numstr = sprintf("%d", $i);
	$newuser = sprintf("%-".$maxlen."s", $base);
	substr($newuser, -length($numstr), length($numstr), $numstr);
	$newuser =~ s/ //g;
        $i++;
    }

    $i >= 10**($maxlen-1) and return undef;
    return $newuser;
}

################################################################################
sub user_exists {
    my ($user, $pwdfile) = @_;

    my $retval = 0;
    my $i;
    my $buffer;

    open my $handle, "<", $pwdfile;
    while (sysread($handle, $buffer, 65536, 0)) {
	if ($buffer =~ /(^|\n)${user}:/) {
	    $retval = 1;
	    last;
	}
    }
    close $handle;
    return $retval;
}

################################################################################
sub usage {
    printf( 
"Usage: ".basename($PROGRAM_NAME)." [-h] [-d] [-mM] [-t] [basename]

Generate a new unused username suitable for assigning to a new account.

basename    If specified this will be used as the basis of the new username

            The basename will be stripped of all characters other than letters 
            and digits, and it will be truncated to no more than the maximum 
            allowed username length (8 by default, but settable with the 
            -m option).
            In the event that a user already exists by the given name a number 
            will be inserted at the end of the username and sequentially 
            increased until an available name is found.
            If basename is not specified, then \"user\" will be used as the base.
-h          Print these instructions.
-d          Enable debug messages.
-mM         Set maximum allowed username length to M (default is 8).
-t          Perform test code.  You must have a suitable passwd file in the 
            current directory in order to run this.
");
}
